import dataclasses
from typing import Annotated, List, Optional

from fastapi import APIRouter, HTTPException
from openassistants.core.assistant import Assistant
from openassistants.data_models.chat_messages import OpasMessage
from openassistants.functions.base import BaseFunction
from openassistants.utils.async_utils import last_value
from pydantic import BaseModel, Field

from openassistants_fastapi.utils.sse import AsyncStreamVersion, sse_json_patch


@dataclasses.dataclass
class RouteAssistants:
    assistants: dict[str, Assistant]


def create_router(route_assistants: RouteAssistants) -> APIRouter:
    v1alpha_router = APIRouter(
        prefix="/v1alpha",
    )

    class ChatRequest(BaseModel):
        messages: Annotated[List[OpasMessage], Field(min_items=1)]
        stream: bool = False
        autorun: Annotated[
            bool, Field(description="automatically run identified function")
        ] = True
        force_select_function: Optional[str] = None

    class ChatResponse(BaseModel):
        messages: Annotated[List[OpasMessage], Field(min_items=1)]

    @v1alpha_router.post("/assistants/{assistant_id}/chat")
    async def chat(
        assistant_id: str,
        body: ChatRequest,
    ):
        if assistant_id not in route_assistants.assistants:
            raise HTTPException(status_code=404, detail="assistant not found")

        assistant = route_assistants.assistants[assistant_id]

        async def stream() -> AsyncStreamVersion[ChatResponse]:
            async for version in assistant.run_chat(
                body.messages, body.autorun, body.force_select_function
            ):
                yield ChatResponse(messages=version)

        if body.stream:
            return sse_json_patch(stream())
        else:
            return await last_value(stream())

    @v1alpha_router.get("/libraries/{assistant_id}/functions")
    async def get_assistant_functions(
        assistant_id: str,
    ) -> List[BaseFunction]:
        if assistant_id not in route_assistants.assistants:
            raise HTTPException(status_code=404, detail="assistant not found")

        return await route_assistants.assistants[assistant_id].get_all_functions()

    @v1alpha_router.get("/libraries/{assistant_id}/functions/{function_id}")
    async def get_function_from_assistant(
        assistant_id: str,
        function_id: str,
    ) -> BaseFunction:
        if assistant_id not in route_assistants.assistants:
            raise HTTPException(status_code=404, detail="assistant not found")
        function_libraries = route_assistants.assistants[
            assistant_id
        ].function_libraries

        for library in function_libraries:
            if (func := await library.aread(function_id)) is not None:
                return func

        raise HTTPException(status_code=404, detail="function not found")

    return v1alpha_router



import inspect
import logging
from typing import Any, Dict, Optional

import jsonpatch  # type: ignore
from fastapi.encoders import jsonable_encoder
from openassistants.utils.async_utils import AsyncStreamVersion
from pydantic import BaseModel
from sse_starlette import EventSourceResponse, ServerSentEvent

logger = logging.getLogger(__name__)


class _SSEJSONPatch(BaseModel):
    patch: Optional[Any] = None
    error: Optional[Any] = None


async def _patch_iter(src: AsyncStreamVersion) -> AsyncStreamVersion[str]:
    if not inspect.isasyncgen(src):
        raise TypeError("src must be an async generator")
    num_sent = 0
    old_dict: Dict[str, Any] = {}
    try:
        async for new in src:
            new_dict = jsonable_encoder(new)
            patch = jsonpatch.JsonPatch.from_diff(old_dict, new_dict)
            old_dict = new_dict
            yield _SSEJSONPatch(patch=patch.patch).json()
            num_sent += 1
    except Exception as e:
        logger.exception("Error while streaming response")
        yield _SSEJSONPatch(error=dict(code=None, detail=str(e))).json()
    finally:
        logger.info(f"closing SSE JSON Patch stream. Sent {num_sent} messages.")


def sse_json_patch(
    src: AsyncStreamVersion,
) -> EventSourceResponse:
    return EventSourceResponse(
        content=_patch_iter(src),
        ping_message_factory=lambda: ServerSentEvent(),
    )
